#!/usr/bin/perl -w
#
#  Invoke as 'perl tlsort.pl PROGRAM PROGRAM PROGRAM <timeline.xml
#  >timeline.html' to generate a timeline containing the named
#  PROGRAMs. '-nochain' suppresses the entry chaining ("next" links),
#  '-nolinks' suppresses the link line at the bottom.
#

use strict;

my @PROGRAMS;
my $opt_chain_links = 1;
my $opt_links = 1;
my %html_tag_map = ( tt => "tt",
                     file => "tt",
                     em => "em",
                     b => "b" );

foreach (@ARGV) {
    if ($_ eq '-nochain') {
        $opt_chain_links = 0;
    } elsif ($_ eq '-nolinks') {
        $opt_links = 0;
    } elsif ($_ =~ /^-/) {
        die "invalid option '$_'";
    } else {
        push @PROGRAMS, $_;
    }
}

my %entries;
my %intros;
my $main_intro;
my %urls;
my $current = undef;
my %years;
while (<STDIN>) {
    while (/<[^>]*$/ && defined (my $line = <STDIN>)) { $_ .= $line }
    s/<!--.*?-->//g;
    if (m{<program\s}) {
        $current = undef;
        if (m{\bname="(.*?)"}) {
            my $pn = $1;
            if (grep {lc($_) eq lc($pn)} @PROGRAMS) { $current = $pn; }
            if (m{\burl="(.*?)"}) {
                $urls{$pn} = $1;
            }
        }
    } elsif (m{</program>}) {
        $current = undef;
    } elsif (m{<entry\s+}) {
        while (!m{</entry>} && defined (my $line = <STDIN>)) { $_ .= $line }
        push @{$entries{$current}}, $_
            if defined($current);
    } elsif (m{<intro>}) {
        while (!m{</intro>} && defined (my $line = <STDIN>)) { $_ .= $line };
        if (defined($current)) {
            $intros{$current} = $_
        }
    } elsif (m{<covertext>}) {
        while (!m{</covertext>} && defined (my $line = <STDIN>)) { $_ .= $line };
        $main_intro = $_;
    }
}

# Validate it.
foreach (keys %entries) {
    validate_chronological_order ($_, @{$entries{$_}});
}

# Now, %entries contains N lists, each in chronological order.
my %index;
foreach (keys %entries) { $index{$_} = 0 }

# Generate links
my %link_names;
if ($opt_chain_links) {
    generate_links();
}

output_init ();
while (1) {
    # Look for oldest entry in all lists
    my $entry_key = undef;
    my $entry_index = undef;
    my $entry_date = undef;
    foreach (keys %entries) {
        foreach my $x ($index{$_} .. $#{$entries{$_}}) {
            my %data = extract_params ($entries{$_}->[$x]);
            if (exists $data{'date'}) {
                if (!defined($entry_date) || $data{'date'} lt $entry_date) {
                    $entry_key = $_;
                    $entry_index = $x;
                    $entry_date = $data{'date'};
                }
                last;
            }
        }
    }

    # not found anything -> only entries with unknown date remain
    last if !defined $entry_date;

    # okay, output this area
    foreach ($index{$entry_key} .. $entry_index) {
        output_entry ($entries{$entry_key}->[$_], $entry_key, $_);
    }
    $index{$entry_key} = $entry_index+1;
}

# emit rest
foreach (sort keys %entries) {
    next if $index{$_} > $#{$entries{$_}};
    foreach my $x ($index{$_} .. $#{$entries{$_}}) {
        output_entry ($entries{$_}->[$x], $_, $x);
    }
}
output_end();

exit 0;

################################## Input ##################################

sub validate_chronological_order {
    my $name = shift;
    my $date = undef;
    foreach (@_) {
        my %things = extract_params($_);
        if (exists $things{"date"}) {
            if (defined $date && $things{"date"} lt $date) {
                print STDERR "$name: non-chronological order: $date vs. $things{'date'}\n";
            }
            $date = $things{'date'};
            if ($date =~ m|^(\d+)/|) { $years{$1} = 1 }
        }
    }
}

sub extract_params {
    my @data;
    my $text = shift;
    if ($text =~ m{<entry\s*([^>]+?)>}) {
        my $data = $1;
        while ($data =~ m{(\w+)="(.*?)"}g) { push @data, $1, $2; }
    }
    @data;
}

################################ Tag Parser ###############################

my $lex_text;
my $lex_token;                  # token type ('open', 'close', 'text')
my $lex_value;                  # token name, or text
my %lex_attribs;                # for open tags, the attributes

sub lex {
    while (1) {
        if ($lex_text =~ m{\G<!--.*?-->}sgc) {
            # empty
        } elsif ($lex_text =~ m{\G</(\w+)>}sgc) {
            # close tag
            $lex_token = 'close';
            $lex_value = $1;
            return $lex_token;
        } elsif ($lex_text =~ m{\G<(\w+)\s*(.*?)>}sgc) { 
            # open tag
            $lex_token = 'open';
            $lex_value = $1;
            %lex_attribs = ();
            my $attribs = $2;
            while ($attribs =~ m{(\w+)="(.*?)"}sgc) { $lex_attribs{$1} = $2; }
            return $lex_token;
        } elsif ($lex_text =~ m{\G([^<]+)}sgc) {
            # close tag
            $lex_token = 'text';
            $lex_value = $1;
            return $lex_token;
        } elsif ($lex_text =~ m{\G\Z}sgc) {
            return ($lex_token = 'eof');
        } else {
            die "parse error at '" . substr($lex_text, pos($lex_text), 30) . "'";
        }
    }
}

sub parse {
    # A document is a list of
    # - text [cdata]
    # - ref [sublist, containing: name, reference to attribs, reference to contents]

    my @list = ();
    while (1) {
        if ($lex_token eq 'text') {
            push @list, $lex_value
                unless @list==0 && $lex_value !~ /\S/;
            lex ();
        } elsif ($lex_token eq 'open') {
            my %sub_attrs = %lex_attribs;
            my @sublist = ($lex_value, \%sub_attrs);
            lex ();
            push @sublist, parse ();
            if ($lex_token ne 'close' || $lex_value ne $sublist[0]) {
                die "expected '</$sublist[0]>', got $lex_token";
            }
            push @list, \@sublist;
            lex ();
        } else {
            pop @list
                if @list && !ref($list[-1]) && $list[-1] !~ /\S/;
            return \@list;
        }
    }
}

sub parse_document {
    $lex_text = shift;
    lex ();
    my $rv = parse ();
    if ($lex_token ne 'eof') {
        die "expected end-of-file, got $lex_token";
    }
    $rv;
}

################################## Output #################################

my $output_indent;

sub output_init {
    $output_indent = 0;
    output_line ('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">');
    output_tag ("<html>", "<head>");
    output_line ('<meta http-equiv="Content-Type" value="text/html; charset=us-ascii">');
    output_line ('<link rel="stylesheet" type="text/css" href="timeline.css">');
    output_tag ('</head>', '<body bgcolor="#ffffff" text="#00000" link="#000080" vlink="#000040">');
    output_line ('<h1>VGA Planets Timeline</h1>');
    output_line ('<p>This timeline covers the following programs:</p>');
    output_tag ('<ul>');
    foreach (sort keys %entries) {
        my $intro = '';
        if (exists $intros{$_}) {
            my $doc = parse_document ($intros{$_});
            $intro = '<br>' . extract_tree_cdata(@$doc);
        }
        if ($opt_chain_links) {
            my $a = make_anchor($_);
            output_line ("<li><a href=\"#" . link_name("${a}", "0") ."\">$_</a>$intro</li>");
        } else {
            output_line ("<li>$_$intro</li>");
        }
    }
    output_tag ('</ul>');
    if (defined $main_intro) {
        my $doc = parse_document ($main_intro);
        output_line ("<p>" . extract_tree_cdata(@$doc) . "</p>");
    }
    output_tag('<table align="center" border><tr><td bgcolor="#aaaaff">');
    my $prefix = "By Year: ";
    foreach (sort keys %years) {
        output_line("$prefix<a href=\"#$_\">$_</a>");
        $prefix = "| ";
    }
    output_tag ('</table>');
    output_line ('<p></p>');         # leave some room
    output_tag ('<table border>');
}

sub output_end {
    my @times = times();
    output_tag ('</table>');
    if ($opt_links) {
        output_line ('<center class="tbc">to be continued...</center>');
        output_line ('<hr>');
        output_tag ('<center>');
        output_line ('[ <a href="vgapage.html">My Planets Page</a>');
        foreach (sort keys %entries) {
            next if not exists $urls{$_};
            output_line("| <a href=\"$urls{$_}\">$_</a>");
        }
        output_line ("]");
        output_tag ('</center>');
        output_line ("<hr>");
    }
    output_tag ('<address>');
    output_line ('This file is maintained by <a href="mailto:Streu@gmx.de">Stefan Reuther</a>.',
                 'It was generated automatically from XML input. Unlike those stupid Java',
                 'things, a simple Perl script needed only ' . ($times[0]+$times[1]) . ' seconds to generate it.');
    output_tag ('</address>', '</body>', '</html>');
}

sub output_line {
    my $indent = " " x $output_indent;
    foreach (@_) {
        my $x = $_;
        $x =~ s/^/$indent/g;
        print "$x\n";
    }
}

sub output_tag {
    foreach my $tag (@_) {
        if ($tag =~ m{^</}) {
            --$output_indent;
            print " " x $output_indent, $tag, "\n";
        } else {
            print " " x $output_indent, $tag, "\n";
            ++$output_indent;
        }
    }
}

sub extract_tree_cdata {
    my $output;
    foreach (@_) {
        if (ref) {
            $output .= extract_tree_cdata (@{$_->[2]})
                if @{$_->[2]}; # I don't know why, but this is needed
        } else {
            $output .= $_;
        }
    }
    $output;
}

sub extract_tree_html {
    my $output;
    foreach (@_) {
        if (ref) {
            my $tag;
            if (exists $html_tag_map{$_->[0]}) { $tag = $html_tag_map{$_->[0]}; }
            $output .= "<$tag>" if defined $tag;
            $output .= extract_tree_cdata (@{$_->[2]})
                if @{$_->[2]}; # I don't know why, but this is needed
            $output .= "</$tag>" if defined $tag;
        } else {
            $output .= $_;
        }
    }
    $output;
}

############################### Entry Output ##############################

my $output_current_year;

sub output_entry {
    my $entry = shift;
    my $program = shift;
    my $index = shift;
    my $entry_ptr = parse_document ($entry);
    if (!ref($entry_ptr->[0]) || $entry_ptr->[0]->[0] ne 'entry') { die "malformed document"; }

    my $version = (exists $entry_ptr->[0][1]{'version'} ? '&nbsp;' . $entry_ptr->[0][1]{'version'} : "");
    my $date = (exists $entry_ptr->[0][1]{'date'} ? '<br>' . $entry_ptr->[0][1]{'date'} : "");
    if (exists $entry_ptr->[0][1]{'date'} && $entry_ptr->[0][1]{'date'} =~ /^(\d\d\d\d)/) {
        if (!defined $output_current_year || $output_current_year ne $1) {
            $output_current_year = $1;
            output_tag ('<tr class="divi">');
            output_tag ('<td colspan="2">');
            output_line ("<a name=\"$output_current_year\">$output_current_year</a>");
            output_tag ("</td>", "</tr>");
        }
    }

    my $class = make_anchor($program);
    output_tag ('<tr class="'.$class.'" valign="top">');
    if ($opt_chain_links) {
        output_line ("<td class=\"pn\"><a name=\"" . link_name($class, $index) . "\">$program$version</a>$date</td>");
#        output_line ("<td class=\"pn\"><a name=\"" . link_name($class, $index) . "\">$program$version</a>$date"
#                     . "<br>" . link_name($class, $index) . "</td>");
    } else {
        output_line ("<td class=\"pn\">$program$version$date</td>");
    }
    output_tag ("<td>");

    my $in_ul = 0;
    foreach (@{$entry_ptr->[0][2]}) {
        if (ref) {
            my $tag = $_->[0];
            if ($tag eq 'feature' || $tag eq 'note' || $tag eq 'change' || $tag eq 'fix') {
                if (!$in_ul) {
                    output_tag ("<ul>");
                    $in_ul = 1;
                }
                output_tag ("<li class=\"$tag\">");
                output_line (extract_tree_html (@{$_->[2]}));
                output_tag ("</li>");
            } else {
                # die "unknown tag $tag";
            }
        } else {
            # ignore. Usually just whitespace.
        }
    }

    my $next = $opt_chain_links && $index < $#{$entries{$program}};
    my $prev = $opt_chain_links && $index > 0;
    if ($in_ul) {
        output_tag ("</ul>");
    } elsif (!$next && !$prev) {
        output_line ("&nbsp;");
    }
    if ($next || $prev) {
        output_tag ("<div class=\"next\" align=\"right\">");
        if ($prev) {
            output_line ("<a href=\"#" . link_name(make_anchor($program), ($index-1)) . "\">Previous</a>");
        }
        if ($next) {
            output_line ("<a href=\"#" . link_name(make_anchor($program), ($index+1)) . "\">Next</a>");
        }
        output_tag ("</div>");
    }

    output_tag ("</td>", "</tr>");
}

sub make_anchor {
    my $x = lc shift;
    $x =~ s/\W//g;
    return $x;
}

################################## Links ##################################

sub generate_links {
    my %used;
    foreach (sort keys %entries) {
        my $a = make_anchor($_);
        foreach my $x (0 .. $#{$entries{$_}}) {
            my %parm = extract_params($entries{$_}[$x]);
            my $suf = defined($parm{version}) ? $parm{version} : defined($parm{date}) ? $parm{date} : "X";
            $suf =~ s/\s.*//;
            $suf =~ s/\W/./g;
            $suf =~ s/\.\.*/./g;
            if (exists $used{$a.$suf}) {
                my $i = 0;
                while (exists $used{$a.$suf.$i}) { ++$i }
                $suf .= $i;
            }
            $link_names{$a."//".$x} = $a.$suf;
            $used{$a.$suf} = 1;
        }
    }
}

sub link_name {
    my $n = shift;
    my $p = shift;
    if (exists $link_names{"$n//$p"}) {
        return $link_names{"$n//$p"};
    } else {
        return $n.$p;
    }
}
